#include <math.h>
#include "xvec.h"

/* 8 jitter points */
static xvar j8[] = {
	-0.334818,  0.435331,
	 0.286438, -0.393495,
	 0.459462,  0.141540,
	-0.414498, -0.192829,
	-0.183790,  0.082102,
	-0.079263, -0.317383,
	 0.102254,  0.299133,
	 0.164216, -0.054399 };

/* 24 jitter points */
static xvar j24[] = {
	 0.030245,  0.136384,
	 0.018865, -0.348867,
	-0.350114, -0.472309,
	 0.222181,  0.149524,
	-0.393670, -0.266873,
	 0.404568,  0.230436,
	 0.098381,  0.465337,
	 0.462671,  0.442116,
	 0.400373, -0.212720,
	-0.409988,  0.263345,
	-0.115878, -0.001981,
	 0.348425, -0.009237,
	-0.464016,  0.066467,
	-0.138674, -0.468006,
	 0.144932, -0.022780,
	-0.250195,  0.150161,
	-0.181400, -0.264219,
	 0.196097, -0.234139,
	-0.311082, -0.078815,
	 0.268379,  0.366778,
	-0.040601,  0.327109,
	-0.234392,  0.354659,
	-0.003102, -0.154402,
	 0.297997, -0.417965 };

xmat	mat0( 0.8,   0,   0, 1, 1.0, 0.5 );
xmat	mat1( 0.8, 0.8, 0.8, 1, 1.0, 0.5 );
xmat	mat2( 0.8, 0.8, 0.8, 1, 1.0, 0.0 );
xmat	mat3( 0.0, 0.8, 0.0, 1, 1.0, 0.5 );

xobj	sphere0( XOBJ_SPHERE,   50,   0, 100, 0,0,0,  100, &mat0 );
xobj	sphere1( XOBJ_SPHERE, -150, -10, 100, 0,0,0,   80, &mat1 );
xobj	sphere2( XOBJ_SPHERE,  -50,-140,   0, 0,0,0,   50, &mat3 );

xobj	plane( XOBJ_PLANE, 100,-100,100, 0,1,-1, 0, &mat2 );

xvec	light0( -1000,  1000, -1000 );
xvec	light1(  1000,   800,  -800 );

xobj *	objs[]		= { &sphere0, &sphere1, &sphere2, &plane };
xvec *	lights[]	= {  &light0, &light1 };

xobj *	exclude		= 0;

static void trace( xray *ray, xvec *c_result, xobj **p_objs, u32 n_objs, xvec **p_lights, u32 n_lights, u32 recursion )
{
	c_result->set( 0, 0, 0 );

	if ( !(recursion--) ) return;

	int		i, j;
	xobj *	obj				= 0;
	xvar	dist, closest	= XVAL_HUGE;

	for ( i = 0; i < n_objs; i++ ) {
		int hit = 0;
		switch ( p_objs[i]->type ) {
			case XOBJ_PLANE: {
				hit = ray->hit_plane( p_objs[i], &dist );
			} break;

			case XOBJ_SPHERE: {
				hit = ray->hit_sphere( p_objs[i], &dist );
			} break;
		}
		if ( hit ) {
			if ( dist < closest ) {
				closest = dist;
				obj = p_objs[i];
			}
		}
	}

	if ( obj ) {
		xvec	isec, norm;

		// intersection
		isec = ray->d * closest + ray->p;

		switch ( obj->type ) {
			case XOBJ_PLANE: {
				norm = obj->n;
			} break;

			case XOBJ_SPHERE: {
				norm = isec - obj->p;
			} break;
		}
		norm.normalize();

		// cast reflections
		xvec	c_refl( 0, 0, 0 );
		if ( obj->mat->reflection > 0.0 ) {
			xray	r_refl;
			r_refl.p = isec;
			r_refl.d = ray->d & norm;
			r_refl.p += r_refl.d * XVAL_EPSILON;
			trace( &r_refl, &c_refl, p_objs, n_objs, p_lights, n_lights, recursion );
		}

		// apply lighting
		xvar	coef_spec = 0, coef_diff = 0;
		xvec	ldir;
		for ( i = 0; i < n_lights; i++ ) {
			xvar diff, spec, c = 0;
			xray r_shadow;
			xvec eye( 0, 0, -1 );

			ldir = *p_lights[i] - isec;
			ldir.normalize();

#if 1
			// diffuse
			diff = ldir ^ norm;
			if ( diff < 0.0 ) diff = 0.0;
			c += diff;
#endif

			// specular
#if 1
			xvec	h;
			h = ldir + eye;
			h.normalize();
			spec = norm ^ h;
			if ( spec < 0.0 ) spec = 0.0;
			coef_spec += pow( spec, 18 );
#endif
			// cast shadows
			exclude = obj;
			r_shadow.p = isec;
			r_shadow.d = ldir;
			for ( j = 0; j < n_objs; j++ ) {
				xvar	dist;
				if ( exclude != p_objs[j] ) {
					if ( r_shadow.hit_sphere( p_objs[j], &dist ) ) {
						c = 0;
					}
				}
			}
			exclude = 0;

			coef_diff += c;
		}
		*c_result = c_refl.lerp( obj->mat->reflection, obj->mat->diffuse ) * coef_diff + (coef_spec * 0.8);
	}
}

void xtracer_render( u32 w, u32 h, u32 *p, void (*update_proc)( void ) )
{
	for ( u32 y = 0, k = 0; y < h; y++, k++ ) {
		for ( u32 x = 0; x < w; x++ ) {

			xvec	c_accum( 0, 0, 0 );
			for ( u32 j = 0; j < 8; j++ ) {
//	u32 j = 0;

				xvec	c_result;
				xray	ray;

				ray.p.set(  (x - (xvar)(w - 1) * 0.5) + 1.2*j8[j << 1],
						   -(y - (xvar)(h - 1) * 0.5) + 1.2*j8[(j << 1) + 1], 0 );
				ray.d.set( 0, 0, 1 );

				trace( &ray, &c_result,
					  objs, sizeof(   objs ) / sizeof(   objs[0] ),
					lights, sizeof( lights ) / sizeof( lights[0] ), 5 );

				c_accum += c_result;
			}
			c_accum *= 1.0 / 8;
			c_accum *= 255;
			p[w * y + x] = c_accum.packed_color();
		}
		if ( k > 20 ) {
			update_proc();
			k = 0;
		}
	}
	update_proc();
}
