#ifndef	__XVEC_H__
#define	__XVEC_H__

#define XVAL_MAX_DISTANCE		1.0e7
#define	XVAL_HUGE				1.0e+17
#define	XVAL_EPSILON			1.0e-10
#define	XVAL_DEPTH_TOLERANCE	1.0e-6

typedef	signed long			s32;
typedef	unsigned long		u32;

typedef	double				xvar;

enum { XOBJ_PLANE = 0, XOBJ_SPHERE };

class xvec {
public:
	xvar	a, b, c;

	// common methods
	xvec() {}
	xvec( xvar x, xvar y, xvar z ) {
		a = x, b = y, c = z;
	}

	void set( xvar x, xvar y, xvar z ) {
		a = x, b = y, c = z;
	}

	xvec & operator + ( xvar x ) {
		xvec r;
		r.a = a + x;
		r.b = b + x;
		r.c = c + x;
		return r;
	}

	xvec & operator + ( xvec &x ) {
		xvec r;
		r.a = a + x.a;
		r.b = b + x.b;
		r.c = c + x.c;
		return r;
	}

	xvec & operator - ( xvec &x ) {
		xvec r;
		r.a = a - x.a;
		r.b = b - x.b;
		r.c = c - x.c;
		return r;
	}

	xvec & operator * ( xvar x ) {
		xvec r;
		r.a = a * x;
		r.b = b * x;
		r.c = c * x;
		return r;
	}

	xvec & operator * ( xvec &x ) {
		xvec r;
		r.a = a * x.a;
		r.b = b * x.b;
		r.c = c * x.c;
		return r;
	}

	xvec & operator / ( xvec &x ) {
		xvec r;
		r.a = a / x.a;
		r.b = b / x.b;
		r.c = c / x.c;
		return r;
	}

	void operator += ( xvar x ) {
		a += x;
		b += x;
		c += x;
	}

	void operator += ( xvec & x ) {
		a += x.a;
		b += x.b;
		c += x.c;
	}

	void operator -= ( xvec & x ) {
		a -= x.a;
		b -= x.b;
		c -= x.c;
	}

	void operator *= ( xvar x ) {
		a *= x;
		b *= x;
		c *= x;
	}

	void operator /= ( xvar x ) {
		a /= x;
		b /= x;
		c /= x;
	}

	// dot product
	xvar operator ^ ( xvec &x ) {
		return a * x.a + b * x.b + c * x.c;
	}

	// reflect
	xvec & operator & ( xvec &n ) {
		xvec r;
		xvar coef = -(*this ^ n) * 2.0;
		r.a = a + n.a * coef;
		r.b = b + n.b * coef;
		r.c = c + n.c * coef;
		return r;
	}

	xvec & lerp( xvar c, xvec & x ) {
		return (*this * c) + (x * (1.0 - c));
	}

	void normalize() {
		xvar l = a * a + b * b + c * c;
		if ( l > 0.0 ) *this *= 1.0 / sqrt( l );
	}

	void clamp( xvar min, xvar max ) {
		if ( a < min ) a = min; else if ( a > max ) a = max;
		if ( b < min ) b = min; else if ( b > max ) b = max;
		if ( c < min ) c = min; else if ( c > max ) c = max;
	}

	u32 packed_color() { clamp( 0, 255 ); return ((u32)a << 16) | ((u32)b << 8) | (u32)c; }
};

class xmat {
public:
	xvec	diffuse;
	xvar	specular;
	xvar	opacity;
	xvar	ior;
	xvar	reflection;

	xmat() {}
	xmat( xvar r, xvar g, xvar b, xvar _opacity, xvar _ior, xvar _reflection ) {
		diffuse.set( r, g, b );
		opacity = _opacity;
		ior = _ior;
		reflection = _reflection;
	}
};

class xobj {
public:
	int		type;
	xvec	p;
	xvec	n;		// normal
	xvar	r, r2;	// radius
	xvar	dist;	// distance from origin
	xmat *	mat;

	xobj() {}
	xobj( int _type, xvar x, xvar y, xvar z, xvar nx, xvar ny, xvar nz, xvar _r, xmat *_mat ) {
		type = _type;
		p.set( x, y, z );
		n.set( nx, ny, nz );
		n.normalize();
		r = _r;
		r2 = r * r;
		dist = sqrt( p ^ p );
		mat = _mat;
	}
};

class xray {
public:
	xvec	p;
	xvec	d;

	int hit_plane( xobj *plane, xvar *depth )
	{
		xvar	NormalDotOrigin, NormalDotDirection;
		xvec	P, D;
		NormalDotDirection = plane->n ^ d;
		if ( fabs( NormalDotDirection ) < XVAL_EPSILON ) return 0;
		NormalDotOrigin = plane->n ^ p;
		*depth = -(NormalDotOrigin + plane->dist) / NormalDotDirection;
		if ( (*depth >= XVAL_DEPTH_TOLERANCE) && (*depth <= XVAL_MAX_DISTANCE) ) return 1;
		return 0;
	}

	int hit_sphere( xobj *sphere, xvar *dist )
	{
		xvar	r2;
		xvar	oc_squared, closest_approach;
		xvar	half_chord, half_chord_squared;

		xvec	origin_to_center = sphere->p - p;

		oc_squared = origin_to_center ^ origin_to_center;
		closest_approach = origin_to_center ^ d;

		if ( (oc_squared >= sphere->r2) && (closest_approach < XVAL_EPSILON) ) return 0;
		half_chord_squared = sphere->r2 - oc_squared + (closest_approach * closest_approach);
		if ( half_chord_squared > XVAL_EPSILON ) {
			half_chord = sqrt( half_chord_squared );
			*dist = closest_approach - half_chord;
			//*d2 = closest_approach + half_chord;
			return 1;
		}
		return 0;
	}
};

void xtracer_render( u32, u32, u32 *, void (*)( void ) );

#endif	// __XVEC_H__
